#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <mysql/mysql.h>
#include <signal.h>
#include <errno.h>

#define BUFFSIZE 8
#define DB_SERV "localhost"
#define DB_USER "root"
#define DB_PASS "kafadar"
#define DB_NAME "HSBS_WSN"
#define MAX_HUM 65.0
#define MIN_HUM 2.0
#define MAX_TEM 35.0
#define MIN_TEM 5.0

MYSQL_RES *res;

MYSQL_ROW row;

struct SensorTable

{
	char *id;
	char *ip;
	char *location;
	int pid;
};

struct SensorTable *SensorInfo = NULL;
unsigned int num_fields;
unsigned int num_sensors;
unsigned long *field_length = NULL;
char quitt;
struct values{
	double val;
	int type;
};

void recordString(char* value, MYSQL *mysql)
{
	int i;
	struct values measure;
	char query[500];
	query[0]=0;
	
	if(value[0] == 'T' || value[0] == 't')
    {
		measure.type = 1;
		for(i = 0; i < num_sensors; i++)
		{
			if(getpid() == SensorInfo[i].pid)
			{
				strcat(query, "INSERT INTO Temperature VALUES (");
				strcat(query, SensorInfo[i].id);
				strcat(query, ",NOW(),");
			}				
		}
		strcat(query, value + 2);
		strcat(query,");");
		measure.val = atof(value+2);
		if (measure.val > MAX_TEM || measure.val < MIN_TEM)
		{
			fprintf(stderr,"Alert!! Temperature:");
			printf("%lf\n", measure.val);
		}
		mysql_query(mysql,query);
	}
	else if(value[0] == 'H' || value[0] == 'h')
	{
		measure.type = 2;
		for(i = 0; i < num_sensors; i++)
		{
			if(getpid() == SensorInfo[i].pid)
			{
				strcat(query, "INSERT INTO Humidity VALUES (");
				strcat(query, SensorInfo[i].id);
				strcat(query, ",NOW(),");
			}				
		}
		strcat(query, value+2);
		strcat(query,");");
		measure.val = atof(value+2);
		
		if (measure.val > MAX_HUM || measure.val < MIN_HUM)
		{
			fprintf(stderr,"Alert!! Humidity:");
			printf("%lf\n", measure.val);
		}
		mysql_query(mysql,query);
	}
}

void getIp(MYSQL *mysql)
{
	int i, j;
	if(mysql_query(mysql, "SELECT * FROM Sensors;"))
		printf("Query error\n");
	
	if(!(res = mysql_use_result(mysql)))
		printf("Storing result error\n");
	
	num_fields = mysql_num_fields(res);
	for(i = 0; row = mysql_fetch_row(res); i++)
	{
		field_length = mysql_fetch_lengths(res);
		if((SensorInfo = (struct SensorTable *) realloc(SensorInfo, (i + 1) * sizeof(struct SensorTable))) == NULL)
		{
			printf("Reallocation problem. Terminating the program...\n");
			exit(1);
		}
		for(j = 0; j < num_fields; j++)
		{
			if(j == 0)
			{
				SensorInfo[i].id = malloc(field_length[j] + 1);
				strcpy(SensorInfo[i].id, row[j]);
				SensorInfo[i].id[field_length[j]] = '\0';
				
			}
			else if(j == 1)
			{
				SensorInfo[i].ip = malloc(field_length[j] + 1);
				strcpy(SensorInfo[i].ip, row[j]);
				SensorInfo[i].ip[field_length[j]] = '\0';
			}
			else if(j == 2)
			{
				SensorInfo[i].location = malloc(field_length[j] + 1);
				strcpy(SensorInfo[i].location, row[j]);
				SensorInfo[i].location[field_length[j]] = '\0';
			}
		}
		num_sensors = i + 1;
	}
	mysql_free_result(res);
}

int main(void) 
{
	int sock;
	struct sockaddr_in echoserver;
	char buffer[BUFFSIZE];
	unsigned int echolen;
	int received = 0;
	int mainPID, lastPID;
	MYSQL mysql;
	int i, j, k;
	int temp_pid;
	int startup = 1;

    if (!mysql_init(&mysql))
	{
		fprintf(stderr,"Failed to initate MySQL\n");
		exit (1);
	}
		
	if (!mysql_real_connect(&mysql,DB_SERV, DB_USER, DB_PASS, DB_NAME, 0, NULL, 0))
	{
		fprintf(stderr,"Unable to connect MySQL Server\n");
		exit (1);
	}
	
	getIp(&mysql);
	mainPID = getpid();
	
    if(mainPID == getpid())
	{
		for (i = 0; i <= num_sensors; i++)
		{
			if (mainPID == getpid())
			{
				temp_pid = fork();
				if (0 == temp_pid)
				{
					if(i != num_sensors)
						SensorInfo[i].pid = getpid();
					else
						lastPID = getpid();
				}
				else
				{
					if(i != num_sensors)
						SensorInfo[i].pid = temp_pid;
					else
						lastPID = temp_pid;
				}
			}
		}
	}
	while((mainPID != getpid()) && (lastPID != getpid()))
	{
		if (!mysql_init(&mysql))
		{
// 			fprintf(stderr,"Failed to initate MySQL\n");
			exit (1);
		}
		
		if (!mysql_real_connect(&mysql,DB_SERV, DB_USER, DB_PASS, DB_NAME, 0, NULL, 0))
		{
// 			fprintf(stderr,"Unable to connect MySQL Server\n");
			exit (1);
		}
		/* Create the TCP socket */
		if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) 
		{
			fprintf(stdout, "Failed to create socket\n");
		}
		/* Construct the server sockaddr_in structure */
		memset(&echoserver, 0, sizeof(echoserver));       /* Clear struct */
		echoserver.sin_family = AF_INET;                  /* Internet/IP */
		for(i = 0; i < num_sensors; i++)
		{
			if(getpid() == SensorInfo[i].pid)
			{
				echoserver.sin_addr.s_addr = inet_addr(SensorInfo[i].ip);  /* IP address of 1st */
				echoserver.sin_port = htons(2000);       /* server port */
				/* Establish connection */
				fprintf(stdout, "Trying to establish a connection with HSBS_Sentinel with ID: %s...\n", SensorInfo[i].id);
				if (connect(sock, (struct sockaddr *) &echoserver, sizeof(echoserver)) < 0)
				{
					fprintf(stdout, "Failed to connect with HSBS_Sentinel %s\n", SensorInfo[i].id);
					startup = 2;
					sleep(5);
				}
				else
					fprintf(stdout, "Receiving from: %s\n", SensorInfo[i].id);
			}
		}
		
		while(1)
		{
			
			if(startup == 2)
			{
				for(i = 0; i < num_sensors; i++)
				{
					if(getpid() == SensorInfo[i].pid)
					{
						k = 0;
						while(1)
						{
							fprintf(stdout, "Trying to establish a connection with HSBS_Sentinel with ID: %s...\n", SensorInfo[i].id);
							if (connect(sock, (struct sockaddr *) &echoserver, sizeof(echoserver)) < 0)
							{
								fprintf(stdout, "Failed to connect with HSBS_Sentinel %s\n", SensorInfo[i].id);
								sleep(5);
							}
							else
							{
								fprintf(stdout, "Receiving from: %s\n", SensorInfo[i].id);
								startup = 1;
								break;
							}
						}
					}
				}
			}
			else if(startup == 1)
			{				
				int bytes = 0;
				bytes = read(sock, buffer, BUFFSIZE-1);
				if (bytes == 0)
				{
					sleep(5);			
					close(sock);
					shutdown(sock, SHUT_RDWR);
					if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) 
					{
						fprintf(stdout, "Failed to create socket\n");
					}
					memset(&echoserver, 0, sizeof(echoserver));       /* Clear struct */
					echoserver.sin_family = AF_INET;                  /* Internet/IP */
					for(i = 0; i < num_sensors; i++)
					{
						if(getpid() == SensorInfo[i].pid)
						{
							echoserver.sin_addr.s_addr = inet_addr(SensorInfo[i].ip);  /* IP address of 1st */
							echoserver.sin_port = htons(2000);       /* server port */
							/* Establish connection */
							k = 0;
							while(1)
							{
								fprintf(stdout, "Trying to establish a connection with HSBS_Sentinel with ID: %s...\n", SensorInfo[i].id);
								if (connect(sock, (struct sockaddr *) &echoserver, sizeof(echoserver)) < 0)
								{
									fprintf(stdout, "Failed to connect with HSBS_Sentinel %s\n", SensorInfo[i].id);
									sleep(5);
								}
								else
								{
									fprintf(stdout, "Receiving from: %s\n", SensorInfo[i].id);
									break;
								}
							}
						}
					}
				}
				received += bytes;
				buffer[bytes] = '\0';        /* Assure null terminated string */
				for(i = 0; i < num_sensors; i++)
				{
					if(getpid() == SensorInfo[i].pid)
					{
						if(bytes == 7)
						{
							fprintf(stdout, "%s	Received from Sensor %s\n", buffer, SensorInfo[i].id);
							recordString(buffer,&mysql);
						}
					}
				}
			}			
		}
		flush(buffer);
		fprintf(stdout, "\n");
		close(sock);
	}
	if(lastPID == getpid())
	{
		while(1)
		{
			scanf("%c", &quitt);
			if((quitt == 'q') || (quitt == 'Q'))
			exit(1);
		}
	}
	if(mainPID == getpid())
	{
		waitpid(lastPID, NULL, 0);
		for(i = 0; i < num_sensors; i++)
		{
			kill(SensorInfo[i].pid, SIGKILL);
		}
		kill(lastPID, SIGKILL);
		exit(0);
	}
}
